"""
Module for management of Minecraft versions
"""

import os
import shutil

import click
from java import undo_renames
from minecraft.source import generate_sources
from minecraft.version import Version
from mint.command import GitCommandError

from shulkr.config import get_config
from shulkr.repo import get_repo


def _commit_version(version: Version) -> None:
	repo = get_repo()
	message_template = get_config().message_template

	commit_msg = message_template.strip().replace('{}', str(version))
	if get_config().undo_renamed_vars and head_has_versions():
		commit_msg += '\n\nRenamed variables reverted'

	repo.git.add('src')

	repo.git.commit(message=commit_msg)


def _tag_version(version: Version) -> None:
	repo = get_repo()

	repo.git.tag(version)


def create_version(version: Version) -> None:
	"""
	Generate the sources for a Minecraft version and commit to the repo

	Args:
		version (Version): Version to create
		undo_renamed_vars (bool): If set, this function will attempt to revert
			any variables that were renamed in the new version
		message_template (str): Template for commit messages ('{}'s will be
			replaced with the version name)
		tag (bool): If set, the commit will be tagged
	"""

	# 1. Generate source code for the current version
	click.secho(f'Generating sources for Minecraft {version}', bold=True)

	repo = get_repo()
	mappings = get_config().mappings
	repo_path = repo.path

	try:
		generate_sources(version, mappings, repo_path)
	except BaseException as e:
		# Undo src/ deletions
		if head_has_versions():
			repo.git.restore('src')
		else:
			path = os.path.join(repo.path, 'src')
			if os.path.exists(path):
				shutil.rmtree(path)

		raise e

	# 2. If there are any previous versions, undo the renamed variables
	if get_config().undo_renamed_vars and head_has_versions():
		click.echo('Undoing renamed variables')
		undo_renames(get_repo().to_gitpython())

	# 3. Commit the new version to git
	click.echo('Committing to git')
	_commit_version(version)

	# 4. Tag
	if get_config().tag:
		_tag_version(version)


def head_has_versions() -> bool:
	"""
	Check if any versions have been generated on the current branch

	Raises:
		e:

	Returns:
		bool: True if at least one version was found on the current branch
	"""

	repo = get_repo()

	try:
		# List tags reachable by HEAD
		repo.git.describe(tags=True)

		# If we made it here, there is at least one tag.
		return True

	except GitCommandError as e:
		if 'fatal: No names found, cannot describe anything.' in e.stderr:
			return False

		raise e


def get_latest_generated_version() -> Version:
	"""
	Get the most recent version commit on the current branch

	Returns:
		Version:
	"""

	if not head_has_versions():
		return None

	repo = get_repo()

	# Get most recent tag reachable by HEAD
	tag_name = repo.git.describe(tags=True)

	return Version.of(tag_name)



from unittest.mock import ANY

import pytest

import mint
from mint.command import GitCommand
from mint.repo import Repo


SUBPROCESS_ANY_ARGS = {
	'cwd': ANY,
	'shell': ANY,
	'check': ANY,
	'capture_output': ANY,
	'text': ANY,
}


@pytest.fixture(autouse=True)
def fs(mocker) -> None:
	mocker.patch('mint.command.subprocess.os.path.exists', return_value=True)
	mocker.patch('mint.command.subprocess.os.path.isfile', return_value=False)


@pytest.fixture(autouse=True)
def subprocess(mocker) -> None:
	mocker.patch('mint.command.subprocess')


@pytest.fixture
def repo() -> Repo:
	return Repo('foo')


@pytest.fixture
def git(repo: Repo) -> GitCommand:
	return GitCommand(repo.path)


class TestGitCommand:
	def test_getattr_calls_subprocess_with_cwd_set_to_repo_path(self, git, repo):
		git.status()

		subprocess_args = {
			**SUBPROCESS_ANY_ARGS,
			'cwd': repo.path
		}

		mint.command.subprocess.run.assert_called_once_with(
			ANY,
			**subprocess_args
		)

	def test_getattr_with_no_arguments_calls_corresponding_git_subcommand(self, git):
		git.status()

		mint.command.subprocess.run.assert_called_once_with(
			'git status',
			**SUBPROCESS_ANY_ARGS
		)

	def test_getattr_with_one_positional_argument_calls_corresponding_git_subcommand_concatenated_with_argument(self, git):
		git.status('src')

		mint.command.subprocess.run.assert_called_once_with(
			'git status src',
			**SUBPROCESS_ANY_ARGS
		)

	def test_getattr_with_one_nonboolean_keyword_argument_calls_corresponding_git_subcommand_concatenated_with_formatted_keyword_argument(self, git):
		git.log(n=3)

		mint.command.subprocess.run.assert_called_once_with(
			'git log -n 3',
			**SUBPROCESS_ANY_ARGS
		)

	def test_getattr_with_one_keyword_argument_set_to_1_sets_the_corresponding_option_to_1(self, git):
		git.log(n=1)

		# It should not be 'git log -n' - it should be 'git log -n 1'
		mint.command.subprocess.run.assert_called_once_with(
			'git log -n 1',
			**SUBPROCESS_ANY_ARGS
		)

	def test_getattr_with_one_keyword_argument_set_to_true_calls_corresponding_git_subcommand_concatenated_with_corresponding_flag(self, git):
		git.log(oneline=True)

		mint.command.subprocess.run.assert_called_once_with(
			'git log --oneline',
			**SUBPROCESS_ANY_ARGS
		)

	def test_getattr_with_one_keyword_argument_set_to_false_calls_corresponding_git_subcommand_concatenated_without_corresponding_flag(self, git):
		git.log(oneline=False)

		mint.command.subprocess.run.assert_called_once_with(
			'git log',
			**SUBPROCESS_ANY_ARGS
		)

	def test_getattr_with_one_positional_argument_and_one_keyword_argument_calls_git_with_the_keyword_argument_before_the_positional_argument(self, git):
		git.log('HEAD', oneline=True)

		mint.command.subprocess.run.assert_called_once_with(
			'git log --oneline HEAD',
			**SUBPROCESS_ANY_ARGS
		)



import os
import shutil
import subprocess

import click

from mint.repo import Repo

from minecraft.version import Version


DECOMPILER_MC_REMOTE_URL = 'https://github.com/hube12/DecompilerMC.git'
YARN_REMOTE_URL = 'https://github.com/FabricMC/yarn.git'


def _setup_decompiler(local_dir: str, remote_url: str) -> Repo:
	if os.path.exists(
		os.path.join(local_dir, '.git')
	):
		# Used cached yarn repo
		return Repo(local_dir)
	else:
		# Clone the yarn repo
		click.echo(f'Cloning {remote_url} into {local_dir}')
		return Repo.clone(remote_url, local_dir)


def _generate_sources_with_yarn(version: Version, path: str) -> None:
	decompiler_path = os.path.join(path, '.yarn')
	decompiler_repo = _setup_decompiler(decompiler_path, YARN_REMOTE_URL)

	click.echo(f'Updating mappings to version {version}')

	# Get latest versions from remote
	decompiler_repo.git.fetch(prune=True)

	decompiler_repo.git.reset('HEAD', hard=True)
	decompiler_repo.git.clean(force=True, d=True)

	# Checkout version branch
	decompiler_repo.git.checkout(f'origin/{version}')

	click.echo('Running decompiler')

	# Generate source code
	p = subprocess.run(
		['./gradlew', 'decompileCFR'],
		stdout=subprocess.DEVNULL,
		stderr=subprocess.PIPE,
		cwd=decompiler_repo.path
	)
	if p.returncode != 0:
		raise Exception(p.stderr.decode())

	click.echo('Moving generated sources')

	src_path = os.path.join(path, 'src')

	# Remove existing top-level destination directory
	if os.path.exists(src_path):
		shutil.rmtree(src_path)

	# Move the generated source code to $repo_dir/src
	shutil.move(
		os.path.join(decompiler_repo.path, 'namedSrc'),
		src_path
	)


def _generate_sources_with_mojang(version: Version, path: str) -> None:
	"""
	Decompiles a version of the Minecraft client with DecompilerMC (which uses
	Mojang's official mappings)

	Args:
		version (Version):

	Raises:
		Exception: If DecompilerMC fails
	"""

	decompiler_path = os.path.join(path, '.DecompilerMC')
	decompiler_repo = _setup_decompiler(decompiler_path, DECOMPILER_MC_REMOTE_URL)

	click.echo('Running decompiler')

	# Decompile client
	p = subprocess.run(
		[
			'python3',
			'main.py',
			'--mcv',
			str(version),
			'-s',
			'client',
			'-c',
			'-f',
			'-q'
		],
		stderr=subprocess.PIPE,
		cwd=decompiler_repo.path
	)
	if p.returncode != 0:
		raise Exception(p.stderr.decode())

	click.echo('Moving generated sources')

	# Sources directory
	dest_src_dir = os.path.join(path, 'src')

	# Remove existing top-level destination directory
	if os.path.exists(dest_src_dir):
		shutil.rmtree(dest_src_dir)

	# Move the generated source code to $dest_dir/src
	shutil.move(
		os.path.join(decompiler_repo.path, 'src', str(version), 'client'),
		dest_src_dir
	)


def generate_sources(version: Version, mappings: str, path: str) -> None:
	if mappings == 'mojang':
		_generate_sources_with_mojang(version, path)

	elif mappings == 'yarn':
		_generate_sources_with_yarn(version, path)

	else:
		raise ValueError(f"Invalid mapping type '{mappings}'")



from __future__ import annotations
import re
import requests
from typing import Dict, List, Optional

MANIFEST_LOCATION = "https://launchermeta.mojang.com/mc/game/version_manifest.json"	 # noqa: E501
EARLIEST_SUPPORTED_VERSION_ID = '19w36a'


class VersionError(Exception):
	pass


class NoSuchVersionError(VersionError):
	pass


class Version:
	def __init__(
		self,
		id: str,
		index: int,
		next: Optional[Version] = None
	) -> None:

		self.id = id
		self._index = index
		self.next = next

	def __str__(self) -> str:
		return self.id

	def __lt__(self, other) -> bool:
		if not isinstance(other, Version):
			return False

		return self._index < other._index

	def __gt__(self, other) -> bool:
		if not isinstance(other, Version):
			return False

		return self._index > other._index

	def to(self, other: Optional[Version], snapshots=True) -> List[Version]:
		"""Iterate over all the versions from this version (inclusive) to
		`other` (exclusive)

		Args:
			other (Optional[Version]): Stop before reaching this version
			include_snapshots (bool, optional): Defaults to True.

		Returns:
			List[Version]: All the versions since this version, before `other`
		"""

		if other < self:
			dots = '...' if snapshots else '..'
			raise ValueError(
				'Version B cannot have been released sooner than version A in'
				+ f' A{dots}B syntax'
			)

		r = []
		curr = self
		while curr is not None and (other is None or curr != other.next):
			if snapshots or isinstance(curr, Release):
				r.append(curr)

			curr = curr.next

		return r

	@staticmethod
	def parse(raw: Dict, index: int) -> Version:
		id = raw['id']
		type_ = raw['type']

		if type_ == 'snapshot':
			return Snapshot(id, index)
		elif type_ == 'release':
			return Release(id, index)
		elif type_ == 'old_alpha':
			return OldAlphaVersion(id, index)
		elif type_ == 'old_beta':
			return OldBetaVersion(id, index)
		else:
			raise Exception(f'Unknown version type: {type_}')

	@staticmethod
	def of(id: Optional[str]) -> Version:
		if id is None:
			return manifest.latest_snapshot

		if id not in manifest.version_for_id:
			raise NoSuchVersionError(f"Unsupported Minecraft version '{id}'")

		return manifest.version_for_id[id]

	@staticmethod
	def pattern(p: str, latest_in_repo: Version = None) -> List[Version]:
		# First, check if the pattern is a range
		m = re.match(r'(.*[^.])?(\.\.\.?)(.*)', p)
		if m is not None:
			a_id, dots, b_id = m.groups()
			snapshots = dots == '...'

			if a_id:
				a = Version.of(a_id)
			else:
				# Get the next version after the latest committed one
				if latest_in_repo:
					a = latest_in_repo.next
					if not snapshots:
						# Find the next release
						while a is not None and not isinstance(a, Release):
							a = a.next
				else:
					raise ValueError('No commits from which to derive current version')

			if a is None:
				# No more versions or releases to generate
				return []

			if b_id:
				b = Version.of(b_id)
			else:
				# Get all the commits since `a`
				b = None

			return a.to(b, snapshots)

		else:
			# Then, try to treat it as a version
			return [Version.of(p)]

	@staticmethod
	def patterns(
		patterns: List[str],
		latest_in_repo: Version = None
	) -> List[Version]:

		# Parse versions
		versions = set()
		for pattern in patterns:
			if pattern.startswith('-'):
				versions -= set(Version.pattern(pattern[1:], latest_in_repo))
			else:
				versions |= set(Version.pattern(pattern, latest_in_repo))

		# Sort
		versions = list(versions)
		versions.sort()

		return versions


class Release(Version):
	@staticmethod
	def of(id: Optional[str]) -> Release:
		if id is None:
			return manifest.latest_release

		else:
			version = Version.of(id)
			if not isinstance(version, Release):
				raise ValueError('Version is not a release')

			return version


class Snapshot(Version):
	pass


class OldVersion(Version):
	pass


class OldAlphaVersion(Version):
	pass


class OldBetaVersion(Version):
	pass


class Manifest:
	def __init__(
		self,
		versions: List[Version],
		earliest_release: Release,
		earliest_snapshot: Version,
		latest_release: Release,
		latest_snapshot: Version
	) -> None:

		self.versions = versions
		self.version_for_id = {version.id: version for version in versions}
		self.earliest_release = earliest_release
		self.earliest_snapshot = earliest_snapshot
		self.latest_release = latest_release
		self.latest_snapshot = latest_snapshot

	def __iter__(self) -> Version:
		return self.versions

	@staticmethod
	def parse(
		raw: Dict,
		earliest_supported_version_id: Optional[str]
	) -> Manifest:

		# Reverse to get ascending chronological order
		reversed_versions = reversed(raw['versions'])

		current_version_supported = False
		earliest_release = None
		earliest_snapshot = None
		versions = []
		version_for_id = {}
		prev = None

		for i, raw_version in enumerate(reversed_versions):
			id = raw_version['id']
			if id == earliest_supported_version_id:
				current_version_supported = True

			if not current_version_supported:
				continue

			version = Version.parse(raw_version, i)
			versions.append(version)
			version_for_id[id] = version

			if prev is not None:
				prev.next = version
			prev = version

			if earliest_snapshot is None:
				earliest_snapshot = version
			if earliest_release is None and isinstance(version, Release):
				earliest_release = version

		latest_release = version_for_id[raw['latest']['release']]
		latest_snapshot = version_for_id[raw['latest']['snapshot']]

		return Manifest(
			versions,
			earliest_release,
			earliest_snapshot,
			latest_release,
			latest_snapshot
		)


def load_manifest(
	raw: Optional[Dict] = None,
	earliest_supported_version_id: Optional[str] = EARLIEST_SUPPORTED_VERSION_ID
):

	global manifest

	if raw is None:
		raw = requests.get(MANIFEST_LOCATION).json()
	manifest = Manifest.parse(raw, earliest_supported_version_id)


def clear_manifest():
	global manifest

	manifest = None


manifest = None



import os

from shulkr.config import Config, init_config


class TestConfig:
	def test_save_opens_config_file_for_writing(self, mocker):
		# 1. Create a new configuration
		config = Config(
			repo_path='foo',
			mappings='mojang',
			message_template='{}',
			tag=True,
			undo_renamed_vars=False
		)

		# 2. Mock open()
		open_ = mocker.patch('shulkr.config.open')

		# 3. Try to save the configuration to the disk
		config.save()

		# 4. Make sure open() was called currectly
		config_path = os.path.join('foo', '.shulkr')
		open_.assert_called_once_with(config_path, 'w+')


def test_init_config_creates_new_configuration_with_provided_arguments_if_config_file_is_not_found(mocker, empty_repo):
	# 1. Spy on Config constructor
	Config_ = mocker.patch('shulkr.config.Config')

	# 2. Patch os.path.exists to return False
	mocker.patch('shulkr.config.os.path.exists', return_value=False)

	# 3. Call init_config
	init_config(
		repo_path='foo',
		mappings='mojang',
		message_template='{}',
		tag=True,
		undo_renamed_vars=False
	)

	# 4. The Config constructor should have been called with the specified
	# path and mappings
	Config_.assert_called_once_with('foo', 'mojang', '{}', True, False)


def test_init_config_commits_config_when_creating_new_one(mocker, empty_repo):
	# 1. Stub out Config class
	mocker.patch('shulkr.config.Config')

	# 2. Mock 'git commit'
	mocker.patch.object(empty_repo.git, 'commit')

	# 3. No existing config should be found
	mocker.patch('shulkr.config.os.path.exists', return_value=False)

	# 4. Call init_config
	init_config(
		repo_path='foo',
		mappings='mojang',
		message_template='{}',
		tag=True,
		undo_renamed_vars=False
	)

	# 5. "git commit --message 'add .shulkr' .shulkr"
	empty_repo.git.add.assert_called_once_with('.shulkr')
	empty_repo.git.commit.assert_called_once_with(message='add .shulkr')


def test_init_config_loads_existing_config_if_config_file_is_found(mocker):
	# 1. Stub out the Config constructor
	Config_ = mocker.patch('shulkr.config.Config')

	# 2. Add a fake config
	# 2a. Patch os.path.exists to return True
	mocker.patch('shulkr.config.os.path.exists', return_value=True)

	# 2b. Stub out open()
	mocker.patch('shulkr.config.open')

	# 2c. Patch toml.load to return a dummy config file
	raw_config = {
		'mappings': 'yarn',
		'message': 'Minecraft {}',
		'tag': False,
		'experimental': {
			'undo_renamed_vars': True
		}
	}
	mocker.patch('shulkr.config.toml.load', return_value=raw_config)

	# 3. Call init_config
	init_config(
		repo_path='foo',
		mappings='mojang',
		message_template='{}',
		tag=True,
		undo_renamed_vars=False
	)

	# 4. The Config constructor should have been called with the path and
	# mappings from the existing config
	Config_.assert_called_once_with(
		repo_path='foo',
		mappings='yarn',
		message_template='Minecraft {}',
		tag=False,
		undo_renamed_vars=True
	)


def test_init_config_returns_true_if_config_file_is_found(mocker):
	# 1. Mock dependencies
	mocker.patch('shulkr.config.Config')
	mocker.patch('shulkr.config.os.path.exists', return_value=True)
	mocker.patch('shulkr.config.open')
	raw_config = {
		'mappings': 'yarn',
		'message': 'Minecraft {}',
		'tag': False,
		'experimental': {
			'undo_renamed_vars': True
		}
	}
	mocker.patch('shulkr.config.toml.load', return_value=raw_config)

	# 2. Call init_config
	result = init_config(
		repo_path='foo',
		mappings='mojang',
		message_template='{}',
		tag=True,
		undo_renamed_vars=False
	)

	# 3. The result should be True
	assert result


def test_init_config_returns_false_if_config_file_is_not_found(mocker, empty_repo):
	# 1. Mock dependencies
	mocker.patch('shulkr.config.Config')
	mocker.patch('shulkr.config.os.path.exists', return_value=False)

	# 2. Call init_config
	result = init_config(
		repo_path='foo',
		mappings='mojang',
		message_template='{}',
		tag=True,
		undo_renamed_vars=False
	)

	# 3. The result should be False
	assert not result



import os
import subprocess
from typing import Any, Dict, List


class GitCommandError(Exception):
	def __init__(self, command: str, stderr: str, *args: object) -> None:
		super().__init__(*args)

		self.command = command
		self.stderr = stderr

	def __str__(self) -> str:
		return f'{self.command}:\n{self.stderr}'


class GitCommand:
	"""
	Context for running git commands

	Sample usage:
		git = GitCommand(PATH_TO_REPO)
		if not git.status(porcelain=True):
			git.commit(message='empty commit', allow_empty=True)

	Currently, options that require an '=' between the key and the value must be
	supplied as positional arguments:
		git.log('--format="%B"')
	"""

	def __init__(self, path: str = None) -> None:
		if path is None:
			path = os.getcwd()

		self._path = path

	def _run_command(self, command: List[str]) -> str:
		try:
			proc = subprocess.run(
				' '.join(command),
				cwd=self._path,
				shell=True,
				check=True,
				capture_output=True,
				text=True
			)

			return proc.stdout.strip()

		except subprocess.CalledProcessError as e:
			raise GitCommandError(
				' '.join(command),
				e.stderr
			)

	def __getattr__(self, name: str):
		"""
		Return the specified git subcommand as a function

		Args:
			name (str): Name of the subcommand
		"""

		def func(*args, **kwargs):
			subcommand = name.replace('_', '-')

			command = GitCommand._raw_command(
				subcommand,
				args,
				kwargs
			)
			return self._run_command(command)

		return func

	@staticmethod
	def _format_value(value: Any):
		if not isinstance(value, str):
			value = str(value)

		if ' ' in value:
			escaped_value = value.replace("'", "\\'")
			return f"'{escaped_value}'"
		else:
			return value

	@staticmethod
	def _format_option(key: str, value: Any) -> List[str]:
		option = key.replace('_', '-')
		prefix = '-' if len(option) == 1 else '--'

		if value is True:
			return [f'{prefix}{option}']

		elif value is False:
			return []

		else:
			# Non-boolean value
			return [f'{prefix}{option}', GitCommand._format_value(value)]

	@staticmethod
	def _raw_command(
		subcommand: str,
		args: List[Any],
		kwargs: Dict[str, Any]
	) -> List[str]:

		options = [
			token for key, value in kwargs.items()
			for token in GitCommand._format_option(key, value)
		]
		args = [GitCommand._format_value(arg) for arg in args]
		return ['git', subcommand, *options, *args]



import os


def test_when_running_with_yarn_repo_only_contains_git_gitignore_yarn_and_src_directories(run_yarn):
	assert set(os.listdir(run_yarn.repo_path)) == set(['.git', '.gitignore', '.shulkr', '.yarn', 'src'])


def test_when_running_with_yarn_src_directory_is_not_empty(run_yarn):
	src_dir = os.path.join(run_yarn.repo_path, 'src')
	assert len(os.listdir(src_dir)) > 0


def test_when_running_with_decompilermc_repo_only_contains_git_gitignore_yarn_client_and_server_directories(run_mojang):
	assert set(os.listdir(run_mojang.repo_path)) == set(['.git', '.gitignore', '.shulkr', '.DecompilerMC', 'src'])


def test_when_running_with_decompilermc_src_directory_is_not_empty(run_mojang):
	src_dir = os.path.join(run_mojang.repo_path, 'src')
	assert len(os.listdir(src_dir)) > 0



from tempfile import TemporaryDirectory

from mint.repo import Repo
import pytest
from minecraft.version import clear_manifest, load_manifest


@pytest.fixture(autouse=True)
def manifest():
	load_manifest()
	yield
	clear_manifest()


@pytest.fixture
def repo(mocker):
	tmp_dir = TemporaryDirectory(prefix='shulkr-test')
	repo = Repo.init(tmp_dir.name)
	mocker.patch('shulkr.repo.repo', repo)

	yield repo

	# tmp_dir is removed when it goes out of scope



from minecraft.version import Version

import shulkr
from shulkr.config import Config
from shulkr.version import create_version, get_latest_generated_version


def test_create_version_calls_generate_sources_with_mappings_from_config_and_correct_version(mocker, config: Config, empty_repo):
	# Set mappings in config
	config.mappings = 'foo'

	# Mock
	mocker.patch('shulkr.version.click')
	mocker.patch('shulkr.version.generate_sources')

	# Call create_version
	version = Version('1.18.1', 0)
	create_version(version)

	# generate_sources() should have been called with the correct arguments
	shulkr.version.generate_sources.assert_called_once_with(version, 'foo', empty_repo.path)


def test_create_version_with_undo_renamed_vars_on_repo_with_no_commits_does_not_call_undo_renames(mocker, config, empty_repo):
	# Mock
	# mocker.patch('shulkr.open', create=True)
	mocker.patch('shulkr.version.click')
	mocker.patch('shulkr.version.generate_sources')
	mocker.patch('shulkr.version.undo_renames')

	# Call create_version
	config.undo_renamed_vars = True
	version = Version('1.18.1', 0)
	create_version(version)

	# Assert that undo_renames was not called
	shulkr.version.undo_renames.assert_not_called()


def test_create_version_with_undo_renamed_vars_on_repo_with_one_commit_calls_undo_renames(mocker, config, nonempty_repo):
	# Mock
	# mocker.patch('shulkr.open', create=True)
	mocker.patch('shulkr.version.click')
	mocker.patch('shulkr.version.generate_sources')
	mocker.patch('shulkr.version.undo_renames')

	# Call create_version
	config.undo_renamed_vars = True
	version = Version('1.18.1', 0)
	create_version(version)

	# Assert that undo_renames was not called
	shulkr.version.undo_renames.assert_called_once()


def test_create_version_with_yarn_mappings_stages_the_src_directory(mocker, config, empty_repo, yarn_mappings):
	mocker.patch('shulkr.version.click')
	mocker.patch('shulkr.version.generate_sources')

	# Call _commit_version
	version = Version('1.18.1', 0)
	create_version(version)

	# src needs to have been staged
	empty_repo.git.add.assert_called_once_with('src')


def test_create_version_with_mojang_mappings_stages_the_src_directory(mocker, config, empty_repo, mojang_mappings):
	mocker.patch('shulkr.version.click')
	mocker.patch('shulkr.version.generate_sources')

	# Call _commit_version
	version = Version('1.18.1', 0)
	create_version(version)

	# client and server need to have been staged
	empty_repo.git.add.assert_called_once_with('src')


def test_create_version_creates_a_commit(mocker, config, empty_repo):
	mocker.patch('shulkr.version.click')
	mocker.patch('shulkr.version.generate_sources')

	# Call _commit_version
	version = Version('1.18.1', 0)
	create_version(version)

	# commit must have been called
	empty_repo.git.commit.assert_called_once()


def test_create_version_replaces_all_brackets_with_the_version(mocker, config, empty_repo):
	mocker.patch('shulkr.version.click')
	mocker.patch('shulkr.version.generate_sources')

	# Call _commit_version
	config.message_template = '{} {}'
	version = Version('1.18.1', 0)
	create_version(version)

	# commit must have been called
	expected_message = config.message_template.replace('{}', str(version))
	empty_repo.git.commit.assert_called_once_with(message=expected_message)


def test_create_version_with_existing_commits_and_undo_renamed_vars_adds_note_to_commit_message(mocker, config, nonempty_repo):
	mocker.patch('shulkr.version.click')
	mocker.patch('shulkr.version.generate_sources')

	# Call _commit_version
	config.undo_renamed_vars = True
	version = Version('1.18.1', 0)
	create_version(version)

	# commit must have been called
	expected_message = f'{version}\n\nRenamed variables reverted'
	nonempty_repo.git.commit.assert_called_once_with(message=expected_message)


def test_create_version_with_tag_calls_git_tag(mocker, config, nonempty_repo):
	mocker.patch('shulkr.version.click')
	mocker.patch('shulkr.version.generate_sources')

	version = Version('1.18.1', 0)
	create_version(version)

	nonempty_repo.git.tag.assert_called_once_with(version)


def test_get_latest_generated_version_with_repo_with_one_version_returns_version(mocker, nonempty_repo):
	mocker.patch('shulkr.version.Version.of')

	# nonempty_repo contains one commit (the snapshot)
	get_latest_generated_version()

	shulkr.version.Version.of.assert_called_once_with('abcdef')



from minecraft.version import Version
import pytest

from shulkr import app


@pytest.fixture
def versions():
	return [
		Version(id='1.18', index=0),
		Version(id='1.19', index=1)
	]


@pytest.fixture(autouse=True)
def mock_all(mocker, versions):
	mocker.patch('shulkr.app.click')
	mocker.patch('shulkr.app.load_manifest')
	mocker.patch('shulkr.app.os')
	mocker.patch('shulkr.app.init_repo')
	mocker.patch('shulkr.app.is_compatible')
	mocker.patch('shulkr.app.init_config')
	mocker.patch('shulkr.app.ensure_gitignore_exists')
	mocker.patch('shulkr.app.Version.patterns', return_value=versions)
	mocker.patch('shulkr.app.get_latest_generated_version')
	mocker.patch('shulkr.app.create_version')


def test_run_loads_version_manifest():
	app.run(
		versions=[],
		mappings='mappings',
		repo_path='path/to/repo',
		message_template='message',
		tags=True,
		undo_renamed_vars=True
	)

	app.load_manifest.assert_called_once_with()


def test_run_calls_init_repo(mocker):
	app.os.path.join.return_value = 'full/path/to/repo'

	app.run(
		versions=[],
		mappings='mappings',
		repo_path='path/to/repo',
		message_template='message',
		tags=True,
		undo_renamed_vars=True
	)

	app.init_repo.assert_called_once_with('full/path/to/repo')


def test_run_with_unsupported_repo_exits_with_error():
	app.is_compatible.return_value = False

	with pytest.raises(SystemExit, match='4'):
		app.run(
			versions=[],
			mappings='mappings',
			repo_path='path/to/repo',
			message_template='message',
			tags=True,
			undo_renamed_vars=True
		)


def test_run_calls_init_config_when_init_repo_returns_false():
	app.init_repo.return_value = False
	app.os.path.join.return_value = 'full/path/to/repo'

	app.run(
		versions=[],
		mappings='mappings',
		repo_path='path/to/repo',
		message_template='message',
		tags=True,
		undo_renamed_vars=True
	)

	app.init_config.assert_called_once_with(
		'full/path/to/repo',
		'mappings',
		'message',
		True,
		True
	)


def test_run_calls_init_config_when_init_repo_returns_true():
	app.init_repo.return_value = True
	app.os.path.join.return_value = 'full/path/to/repo'

	app.run(
		versions=[],
		mappings='mappings',
		repo_path='path/to/repo',
		message_template='message',
		tags=True,
		undo_renamed_vars=True
	)

	app.init_config.assert_called_once_with(
		'full/path/to/repo',
		'mappings',
		'message',
		True,
		True
	)


def test_run_calls_ensure_gitignore_exists():
	app.run(
		versions=[],
		mappings='mappings',
		repo_path='path/to/repo',
		message_template='message',
		tags=True,
		undo_renamed_vars=True
	)

	app.ensure_gitignore_exists.assert_called_once_with()


def test_run_with_version_older_than_latest_version_in_repo_exits_with_error():
	app.get_latest_generated_version.return_value = Version(id='1.18', index=1)

	with pytest.raises(SystemExit, match='3'):
		app.run(
			versions=[Version(id='1.17', index=0)],
			mappings='mappings',
			repo_path='path/to/repo',
			message_template='message',
			tags=True,
			undo_renamed_vars=True
		)


def test_run_with_multiple_versions_calls_create_version_for_each_version(mocker, versions):
	app.run(
		versions=[],
		mappings='mappings',
		repo_path='path/to/repo',
		message_template='message',
		tags=True,
		undo_renamed_vars=True
	)

	app.create_version.assert_has_calls([
		mocker.call(version) for version in versions
	])


def test_run_without_any_versions_exits():
	app.Version.patterns.return_value = []

	with pytest.raises(SystemExit, match='0'):
		app.run(
			versions=[],
			mappings='mappings',
			repo_path='path/to/repo',
			message_template='message',
			tags=True,
			undo_renamed_vars=True
		)



from timeit import timeit

from java import get_renamed_variables


def wrap_in_class(code: str) -> str:
	return """package foo;

	class Foo {
		public static void main(String[] args) {
			%s
		}
	}
	""" % code


def profile_get_renamed_variables():
	SAMPLES = [
		('int x = 0;', 'int x = 0;'),
		('int x = 0;', 'int y = 0;'),
		('int x, y, z;', 'int a, b, c;'),
		('int x; int y; int z;', 'int a; int b; int c;'),
		(
			'if (true) { int x = 0; } else { int x = 0; }',
			'if (true) { int x = 0; } else { int x = 0; }'
		),
		(
			'; '.join([f'int x{i}' for i in range(100)]) + ';',
			'; '.join([f'int x{i}' for i in range(100)]) + ';'
		),
		(
			'; '.join([f'int x{i}' for i in range(100)]) + ';',
			'; '.join([f'int y{i}' for i in range(100)]) + ';'
		)
	]

	total_time = 0.0

	for before, after in SAMPLES:
		print('---------------')
		print(before, end='\n')
		print(after, end='\n')
		before_wrapped = wrap_in_class(before)
		after_wrapped = wrap_in_class(after)
		duration = timeit(
			lambda: get_renamed_variables(before_wrapped, after_wrapped),
			number=100
		)
		total_time += duration
		print(f'{duration}s', end='\n\n')

	print('---------------')
	print(f'{total_time}s')



from mint.repo import Repo


def test_git_repo_initiated(run_yarn):
	# Make sure this is a git repo. If it's not, an error will be thrown.
	Repo(run_yarn.repo_path)


def test_commits_created(run_yarn):
	repo = Repo(run_yarn.repo_path)
	history = repo.git.rev_list('HEAD').splitlines()
	actual = [repo.git.log("--format='%B'", commit, n=1) for commit in history]
	actual.reverse()

	# Calculate expected commit messages from versions
	expected = ['add .shulkr', 'add .gitignore']
	if run_yarn.undo_renamed_vars:
		expected.append(f'version {run_yarn.versions[0]}')
		expected.extend([
			f'version {version}\n\nRenamed variables reverted'
			for version in run_yarn.versions[1:]
		])
	else:
		expected.extend([f'version {version}' for version in run_yarn.versions])

	assert actual == expected


def test_when_running_with_yarn_gitignore_and_src_are_tracked(run_yarn):
	repo = Repo(run_yarn.repo_path)

	# List files and directories that were changed directly under the root
	actual = set(repo.git.ls_tree('HEAD', name_only=True).splitlines())
	expected = set(['.gitignore', '.shulkr', 'src'])
	assert actual == expected


def test_when_running_with_decompilermc_gitignore_src_are_tracked(run_mojang):
	repo = Repo(run_mojang.repo_path)

	# List files and directories that were changed directly under the root
	actual = set(repo.git.ls_tree('HEAD', name_only=True).splitlines())
	expected = set(['.gitignore', '.shulkr', 'src'])
	assert actual == expected


def test_tags_created(run_yarn):
	repo = Repo(run_yarn.repo_path)
	actual = set(repo.git.tag(list=True).splitlines())
	expected = set(run_yarn.versions)
	assert actual == expected


def test_when_running_with_yarn_working_tree_clean(run_yarn):
	repo = Repo(run_yarn.repo_path)
	assert len(repo.git.status(porcelain=True)) == 0


def test_when_running_with_decompilermc_working_tree_clean(run_mojang):
	repo = Repo(run_mojang.repo_path)
	assert len(repo.git.status(porcelain=True)) == 0



from __future__ import annotations
import os

import click
import toml

from shulkr.repo import get_repo


class Config:
	def __init__(
		self,
		repo_path: str,
		mappings: str,
		message_template: str,
		tag: bool,
		undo_renamed_vars: bool
	) -> None:

		self.repo_path = repo_path
		self.mappings = mappings
		self.message_template = message_template
		self.tag = tag
		self.undo_renamed_vars = undo_renamed_vars

	def save(self) -> None:
		"""
		Write this configuration as TOML to the .shulkr file in the
		corresponding shulkr repo
		"""

		raw_config = {
			'mappings': self.mappings,
			'message': self.message_template,
			'tag': self.tag,
			'experimental': {
				'undo_renamed_vars': self.undo_renamed_vars
			}
			# No need to store the repo path (since it is supplied to the CLI
			# and defaults to the CWD)
		}

		config_path = _config_path_for_repo(self.repo_path)
		with open(config_path, 'w+') as config_file:
			toml.dump(raw_config, config_file)


def _config_path_for_repo(repo_path: str) -> str:
	return os.path.join(repo_path, '.shulkr')


def _load_config(repo_path: str) -> Config:
	config_path = _config_path_for_repo(repo_path)
	with open(config_path, 'r') as config_file:
		raw_config = toml.load(config_file)

	return Config(
		repo_path=repo_path,
		mappings=raw_config['mappings'],
		message_template=raw_config['message'],
		tag=raw_config['tag'],
		undo_renamed_vars=raw_config['experimental']['undo_renamed_vars']
	)


def _commit_config() -> None:
	repo = get_repo()

	repo.git.add('.shulkr')
	repo.git.commit(message='add .shulkr')


def _create_config(
	repo_path: str,
	mappings: str,
	message_template: str,
	tag: bool,
	undo_renamed_vars: bool
) -> Config:

	global config

	click.echo('Saving config')

	config = Config(
		repo_path,
		mappings,
		message_template,
		tag,
		undo_renamed_vars
	)
	config.save()
	_commit_config()

	return config


def config_exists(repo_path: str) -> bool:
	return os.path.exists(
		_config_path_for_repo(repo_path)
	)


def init_config(
	repo_path: str,
	mappings: str,
	message_template: str,
	tag: bool,
	undo_renamed_vars: bool
) -> bool:
	"""
	Initialize the config state

	If a .shulkr file exists for the current repo, it will be loaded.
	Otherwise, a new one will be created with the specified mappings.

	Args:
		repo_path (str): _description_
		mappings (str): _description_

	Returns:
		bool: True if a config was loaded. False if a new one was created.
	"""

	global config

	if config_exists(repo_path):
		config = _load_config(repo_path)
		return True
	else:
		config = _create_config(
			repo_path,
			mappings,
			message_template,
			tag,
			undo_renamed_vars
		)
		return False


def clear_config() -> None:
	"""
	Unload the config from memory

	Used in tests
	"""

	global config

	config = None


def get_config():
	return config


config = None



import sys
from typing import List

import click

from shulkr.app import run


@click.command(
	name='shulkr',
	help='Generate multiple versions of the Minecraft source code'
)
@click.help_option(
	'-h',
	'--help',
	is_flag=True,
	help='Show this message and exit'
)
@click.option(
	'--mappings',
	'-p',
	type=click.Choice(['yarn', 'mojang']),
	default='yarn',
	help="Mappings for deobfuscation (defaults to 'yarn')"
)
@click.option(
	'--repo',
	'-p',
	type=click.Path(),
	default='.',
	help='Path to the Minecraft repo (defaults to the current working directory)'
)
@click.option(
	'--message',
	'-m',
	type=str,
	default='version {}',
	help="Commit message template (defaults to 'version {}')"
)
@click.option(
	'--no-tags',
	'-T',
	is_flag=True,
	help='Do not tag commits'
)
@click.option(
	'--undo-renamed-vars',
	'-u',
	is_flag=True,
	help=(
		'Revert local variables that were renamed in new versions to their '
		'original names (experimental)'
	)
)
@click.argument('versions', nargs=-1, type=click.STRING)
def cli(
	versions: List[str],
	mappings: str,
	repo: str,
	message: str,
	no_tags: bool,
	undo_renamed_vars: bool
) -> None:

	tags = not no_tags
	try:
		run(versions, mappings, repo, message, tags, undo_renamed_vars)

	except ValueError as e:
		click.secho(e, err=True, fg='red')
		sys.exit(2)

	except KeyboardInterrupt:
		click.echo('Aborted!', err=True)



from shulkr.repo import init_repo


def test_init_repo_returns_true_if_repo_exists(mocker, empty_repo):
	mocker.patch('shulkr.repo.click')
	mocker.patch('shulkr.repo.Repo')

	assert init_repo(empty_repo.path)


def test_init_repo_returns_false_if_repo_directory_does_not_exist(mocker):
	mocker.patch('shulkr.repo.click')
	mocker.patch('shulkr.repo.Repo', side_effect=FileNotFoundError)

	assert not init_repo('/tmp/does-not-exist')



import pytest

import shulkr
from shulkr.compatibility import is_compatible


@pytest.fixture(autouse=True)
def mock_all(mocker):
	mocker.patch('shulkr.compatibility.config_exists')


def test_is_compatible_with_no_config_and_empty_repo_returns_true(empty_repo):
	shulkr.compatibility.config_exists.return_value = False
	assert is_compatible()


def test_is_compatible_with_no_config_and_nonempty_repo_returns_false(nonempty_repo):
	shulkr.compatibility.config_exists.return_value = False
	assert not is_compatible()


def test_is_compatible_with_config_and_nonempty_repo_returns_true(nonempty_repo):
	shulkr.compatibility.config_exists.return_value = True
	assert is_compatible()



import os
from typing import Optional

import git


def get_blob(repo: git.Repo, commit: Optional[git.Commit], path: str) -> bytes:
	if commit is None:
		p = os.path.join(repo.working_tree_dir, path)
		with open(p, 'r') as f:
			return f.read()

	parts = path.split('/')
	curr = commit.tree

	while len(parts) > 0:
		name = parts.pop(0)
		curr = curr[name]

	return curr.data_stream.read().decode()



from javalang.tree import CompilationUnit, ClassDeclaration, Literal, MethodDeclaration

from java import ast_nodes_equal, filter_ast_node, get_renamed_variables, undo_variable_renames


def wrap_in_class(code: str) -> str:
	return """package foo;

	class Foo {
		public static void main(String[] args) {
			%s
		}
	}
	""" % code


def test_ast_nodes_equal_with_equal_ints_returns_true():
	assert ast_nodes_equal(1, 1, recursive=False)


def test_ast_nodes_equal_recursive_with_equal_ints_returns_true():
	assert ast_nodes_equal(1, 1)


def test_ast_nodes_equal_with_different_ints_returns_true():
	assert not ast_nodes_equal(1, 2, recursive=False)


def test_ast_nodes_equal_recursive_with_different_ints_returns_true():
	assert not ast_nodes_equal(1, 2)


def test_ast_nodes_equal_with_equal_nodes_returns_true():
	node1 = Literal(value='hello')
	node2 = Literal(value='hello')
	assert ast_nodes_equal(node1, node2, recursive=False)


def test_ast_nodes_equal_recursive_with_equal_nodes_returns_true():
	node1 = Literal(value='hello')
	node2 = Literal(value='hello')
	assert ast_nodes_equal(node1, node2)


def test_ast_nodes_equal_with_inequal_nodes_returns_false():
	node1 = Literal(value='hello')
	node2 = Literal(value='bye')
	assert not ast_nodes_equal(node1, node2, recursive=True)


def test_ast_nodes_equal_recursive_with_inequal_nodes_returns_false():
	node1 = Literal(value='hello')
	node2 = Literal(value='bye')
	assert not ast_nodes_equal(node1, node2)


def test_ast_nodes_equal_recursive_with_equal_nodes_wrapped_in_lists_returns_true():
	node1 = Literal(value='hello')
	node2 = Literal(value='hello')
	assert ast_nodes_equal([node1], [node2])


def test_ast_nodes_equal_recursive_with_inequal_nodes_wrapped_in_lists_returns_false():
	node1 = Literal(value='hello')
	node2 = Literal(value='bye')
	assert not ast_nodes_equal([node1], [node2])


def test_ast_nodes_equal_recursive_with_inequal_nodes_and_the_inequal_part_ignored_returns_true():
	# The following doesn't make any sense, but it's just to test our code, calm
	# down
	node1 = Literal(value='hello')
	node2 = Literal(value='bye')
	node3 = Literal(value=node1)
	node4 = Literal(value=node2)
	assert ast_nodes_equal(node3, node4, matches=[(node1, node2, 'value')])


def test_filter_ast_node_with_no_matching_children_returns_empty_list():
	node = Literal(value='hello')
	assert list(filter_ast_node(node, CompilationUnit)) == []


def test_get_renamed_variables_identical_code_with_single_declaration_with_intialization_returns_empty_list():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x = 0;'),
		wrap_in_class('int x = 0;')
	)

	assert renamed_variables == []


def test_get_renamed_variables_identical_code_with_two_declarations_without_initializations_returns_empty_list():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x; int y;'),
		wrap_in_class('int x; int y;')
	)

	assert renamed_variables == []


def test_get_renamed_variables_identical_code_with_one_declaration_with_two_variables_without_initializations_returns_empty_list():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x, y;'),
		wrap_in_class('int x, y;')
	)

	assert renamed_variables == []


def test_get_renamed_variables_single_declaration_with_intialization_returns_the_mapping():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x = 0;'),
		wrap_in_class('int y = 0;')
	)

	assert [v for path, v in renamed_variables] == [[('x', 'y')]]


def test_get_renamed_variables_single_declaration_with_intialization_returns_correct_path():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x = 0;'),
		wrap_in_class('int y = 0;')
	)

	actual = [type(node) for path, v in renamed_variables for node in path]
	expected = [CompilationUnit, str, ClassDeclaration, str, MethodDeclaration, str]
	assert actual == expected


def test_get_renamed_variables_two_declarations_without_intializations_returns_both_mappings():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x; int a;'),
		wrap_in_class('int y; int b;')
	)

	assert [v for path, v in renamed_variables] == [[('x', 'y'), ('a', 'b')]]


def test_get_renamed_variables_two_declarations_with_intializations_returns_both_mappings():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x = 0; int a = 1;'),
		wrap_in_class('int y = 0; int b = 1;')
	)

	assert [v for path, v in renamed_variables] == [[('x', 'y'), ('a', 'b')]]


def test_get_renamed_variables_one_declaration_with_different_intializations_returns_empty_list():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x = 0;'),
		wrap_in_class('int y = 1;')
	)

	assert renamed_variables == []


def test_get_renamed_variables_single_declaration_without_intialization_returns_the_mapping():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x;'),
		wrap_in_class('int y;')
	)

	assert [v for path, v in renamed_variables] == [[('x', 'y')]]


def test_get_renamed_variables_single_declaration_with_one_identical_reference_returns_the_mapping():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x; System.out.println(x);'),
		wrap_in_class('int y; System.out.println(y);')
	)

	assert [v for path, v in renamed_variables] == [[('x', 'y')]]


def test_get_renamed_variables_single_declaration_with_one_different_reference_returns_empty_list():
	renamed_variables = get_renamed_variables(
		wrap_in_class('int x; System.out.println(x);'),
		wrap_in_class('int y; System.out.println(y + "...");')
	)

	assert renamed_variables == []


def test_get_renamed_variables_one_declaration_without_initialization_in_each_branch_of_if_statement_returns_the_mapping():
	renamed_variables = get_renamed_variables(
		wrap_in_class('if (true) { int x; } else { int x; }'),
		wrap_in_class('if (true) { int y; } else { int y; }')
	)

	assert [v for path, v in renamed_variables] == [[('x', 'y')], [('x', 'y')]]


def test_undo_variable_renames_with_one_declaration_and_no_renames_returns_identical_code():
	code = 'int x;'
	renamed_variables = []

	assert undo_variable_renames(code, renamed_variables) == code


def test_undo_variable_renames_with_one_renamed_declaration_updates_code():
	renamed_variables = [(None, [('x', 'y')])]

	assert undo_variable_renames('int y;', renamed_variables) == 'int x;'


def test_undo_variable_renames_with_one_declaration_with_one_renamed_variable_and_one_unchanged_variable_updates_renamed_variable():
	renamed_variables = [(None, [('x', 'y')])]

	assert undo_variable_renames('int a, y;', renamed_variables) == 'int a, x;'


def test_undo_variable_renames_with_one_renamed_declaration_with_one_reference_updates_declaration_and_reference():
	renamed_variables = [(None, [('x', 'y')])]

	assert undo_variable_renames('int y = 0; int a = y;', renamed_variables) == 'int x = 0; int a = x;'


def test_undo_variable_renames_with_one_renamed_declaration_set_to_string_with_same_text_only_updates_identifier():
	renamed_variables = [(None, [('x', 'y')])]

	assert undo_variable_renames('int y = "y";', renamed_variables) == 'int x = "y";'



import os
from unittest.mock import MagicMock

from mint.command import GitCommandError
from mint.repo import Repo
import pytest
from shulkr.config import Config, get_config
from shulkr.repo import get_repo


def create_repo(mocker, path: str):
	repo = MagicMock()

	# Make the repo directory useful for testing shulkr.minecraft.source
	repo.path = os.path.abspath(path)

	class MockGitCommand:
		def add():
			pass

		def checkout():
			pass

		def clean():
			pass

		def commit():
			pass

		def describe():
			pass

		def fetch():
			pass

		def reset():
			pass

		def tag():
			pass

		def rev_parse():
			pass

	repo.git = mocker.create_autospec(MockGitCommand())

	return repo


@pytest.fixture
def config(mocker):
	"""
	Use a fake config

	The advantage of using this fixture over calling init_config is that this
	fixture does not use any operating system resources.
	"""

	config = Config(
		repo_path=os.path.abspath('foo'),
		mappings='mojang',
		message_template='{}',
		tag=True,
		undo_renamed_vars=False
	)
	mocker.patch('shulkr.config.config', config)

	return config


@pytest.fixture
def decompiler(mocker):
	"""
	Mock _setup_decompiler() to return a fake git repo with the current
	path
	"""

	def mocked_setup_decompiler(local_dir: str, _remote_url: str) -> Repo:
		"""Create a fake decompiler subdirectory (.yarn or .DecompilerMC)"""

		# It will be located directly under the shulkr repo directory
		repo = get_repo()
		decompiler_dir = os.path.join(repo.path, local_dir)

		# Create a fake git repo, and set the path property
		return create_repo(mocker, decompiler_dir)

	# Tell the generate_sources functions to use our fake decompiler creator
	mocker.patch(
		'minecraft.source._setup_decompiler',
		new=mocked_setup_decompiler
	)


@pytest.fixture
def empty_repo(mocker, decompiler):
	repo = create_repo(mocker, 'foo')

	# Throw error when `git rev-parse` is called
	def rev_parse(*args, **kwargs):
		raise GitCommandError(
			'git rev-parse...',
			stderr="fatal: ambiguous argument 'HEAD': unknown revision or path not in the working tree."
		)

	mocker.patch.object(repo.git, 'rev_parse', side_effect=rev_parse)

	# Throw error when 'git describe' is called (it will only be called with
	# --tags)
	describe_error = GitCommandError(
		'git describe...',
		stderr='fatal: No names found, cannot describe anything.'
	)
	mocker.patch.object(repo.git, 'describe', side_effect=describe_error)

	# get_repo() will return this value
	mocker.patch('shulkr.repo.repo', repo)

	return repo


@pytest.fixture
def nonempty_repo(mocker, decompiler):
	repo = create_repo(mocker, 'foo')

	# Add a fake commit
	mocker.patch.object(
		repo.git,
		'rev_parse',
		return_value=iter(['9e71573c6ae5a52195274871a679a23379ad1274'])
	)

	# Add a fake tag for that commit (it will only be called with --tags)
	mocker.patch.object(repo.git, 'describe', return_value='abcdef')

	# get_repo() will return this value
	mocker.patch('shulkr.repo.repo', repo)

	return repo


@pytest.fixture
def yarn_mappings(config):
	config = get_config()

	prev_mappings = config.mappings
	config.mappings = 'yarn'

	yield

	config.mappings = prev_mappings


@pytest.fixture
def mojang_mappings(config):
	config = get_config()

	prev_mappings = config.mappings
	config.mappings = 'mojang'

	yield

	config.mappings = prev_mappings


@pytest.fixture
def root_dir():
	script_dir = os.path.dirname(__file__)
	return os.path.realpath(
		os.path.join(script_dir, '..', '..')
	)



import os


class TestRepo:
	def test_init_creates_git_directory(self, repo):
		git_dir = os.path.join(repo.path, '.git')
		assert os.path.exists(git_dir)

	def test_clone_creates_git_directory(self, shallow_cloned_repo):
		git_dir = os.path.join(shallow_cloned_repo.path, '.git')
		assert os.path.exists(git_dir)



from tempfile import TemporaryDirectory
import pytest

from mint.repo import Repo


@pytest.fixture
def tempdir():
	with TemporaryDirectory() as tempdir:
		yield tempdir


@pytest.fixture
def repo(tempdir):
	yield Repo.init(tempdir)


@pytest.fixture
def shallow_cloned_repo(tempdir):
	yield Repo.clone('https://github.com/clabe45/shulkr.git', tempdir, depth=1)



"""
Internal git wrapper for Python

The goal of Mint is to provide a Pythonic interface to run git commands. We
were previously using gitpython, but we are now migrating to Mint because
gitpython's use of git's plumbing commands can lead to unexpected results. To
ease the migration process, Mint's API was designed to be very similar to that
of gitpython.
"""



from __future__ import annotations

import click

from mint.repo import NoSuchRepoError, Repo


def init_repo(repo_path: str) -> bool:
	"""
	Load information about the current shulkr/git repo

	Must be called before get_repo()

	Args:
		repo_path (str): Path to the working tree of the repo

	Returns:
		bool: True if a repo was found, False if a new repo was created
		loaded.
	"""

	global repo

	try:
		repo = Repo(repo_path)

		# The repo already exists
		return True

	except FileNotFoundError:
		click.echo('Initializing git')
		repo = Repo.init(repo_path)

	except NotADirectoryError:
		click.echo('Initializing git')
		repo = Repo.init(repo_path)

	except NoSuchRepoError:
		click.echo('Initializing git')
		repo = Repo.init(repo_path)

	# We created the repo
	return False


def get_repo() -> Repo:
	return repo


repo = None



import os
from typing import Dict, List, Tuple, Union

import git
import javalang
from javalang.ast import Node
from javalang.tokenizer import Identifier, tokenize
from javalang.tree import MemberReference, VariableDeclaration

from java.blob import get_blob


class JavaAnalyzationError(Exception):
	pass


def _ast_nodes_equal(
	node1: Node,
	node2: Node,
	recursive: bool,
	matches: List[Tuple[any, any]],
	_first_level=True
) -> bool:

	# Non-recursive comparison; we're done now that the first node has been
	# processed
	if not recursive and not _first_level:
		return True

	if type(node1) is not type(node2):
		return False

	if type(node1) is list:
		if len(node1) != len(node2):
			return False

		for i in range(len(node1)):
			if not _ast_nodes_equal(node1[i], node2[i], recursive, matches, False):
				return False

		return True

	elif isinstance(node1, Node):
		ignored_attrs = [
			attrs for a, b, attrs in matches if node1 in (a, b) or node2 in (a, b)
		]
		for attr in node1.attrs:
			if attr in ignored_attrs:
				continue

			attr1 = getattr(node1, attr)
			attr2 = getattr(node2, attr)

			if not _ast_nodes_equal(attr1, attr2, recursive, matches, False):
					return False

		return True

	else:
		# if node1 != node2:
			# print('XD', node1, node2, '\n')
		return node1 == node2


def ast_nodes_equal(
	node1: Node,
	node2: Node,
	recursive=True,
	matches: List[Tuple[any, any]] = tuple()
) -> bool:

	return _ast_nodes_equal(node1, node2, recursive, matches)


def ast_paths_equal(
	path1: List[Union[Node, List[Node]]],
	path2: List[Union[Node, List[Node]]],
	matches: List[Tuple[any, any]] = tuple()
) -> bool:
	"""
	Determine if two AST node paths are equal, ignoring other parts of the file

	Args:
		path1 (List[Union[Node, List[Node]]]): [description]
		path2 (List[Union[Node, List[Node]]]): [description]
		matches (List[Tuple[any, any]], optional): List of node pairs to
			consider equal. Defaults to tuple().

	Returns:
		bool: [description]
	"""

	if len(path1) != len(path2):
		return False

	# Skip list entries in the path (because they contain irrelevant parts)
	for i in range(len(path1)):
		if not ast_nodes_equal(path1[i], path2[i], recursive=False, matches=matches):
			return False

	return True


def _filter_ast_node(node: Node, t: type, _path=[]) -> List[Union[Node, str]]:
	if node is None:
		return

	elif type(node) is list:
		for item in node:
			for path, child in _filter_ast_node(item, t, _path=_path):
				yield path, child

	elif isinstance(node, Node):
		if isinstance(node, t):
			yield _path, node

		for attr in node.attrs:
			yield from _filter_ast_node(
				getattr(node, attr),
				t,
				_path=_path + [node, attr]
			)


def filter_ast_node(node: Node, t: type) -> List[Union[Node, str]]:
	return _filter_ast_node(node, t)


def chunk_ast_nodes_by_path(
	nodes: List[Tuple[List, Node]]
) -> List[Tuple[List, List[Node]]]:

	chunks = []
	curr_path = None
	curr_chunk = None
	for path, node in nodes:
		if path != curr_path:
			if curr_path is not None:
				chunks.append((curr_path, curr_chunk))

			curr_path = path
			curr_chunk = []

		curr_chunk.append(node)

	if curr_path is not None:
		chunks.append((curr_path, curr_chunk))

	return chunks


def _have_same_references(
	a_var: str,
	b_var: str,
	a_refs: List[Tuple],
	b_refs: List[Tuple],
	matches: List[Tuple]
) -> bool:

	ar = [
		(path, ref)
		for path, refs in a_refs
		for ref in refs
		if ref.member == a_var
	]

	br = [
		(path, ref)
		for path, refs in b_refs
		for ref in refs
		if ref.member == b_var
	]

	if len(ar) != len(br):
		return False

	for i in range(len(ar)):
		a_ref_path, a_ref_node = ar[i]
		b_ref_path, b_ref_node = br[i]
		ref_match = (a_ref_node, b_ref_node, 'name')

		if not ast_paths_equal(
			a_ref_path,
			b_ref_path,
			matches=matches + [ref_match]
		):
			return False

	return True


def get_renamed_variables(
	source_code: str,
	target_code: str
) -> Dict[str, str]:

	try:
		source_tree = javalang.parse.parse(source_code)
		target_tree = javalang.parse.parse(target_code)
	except javalang.parser.JavaSyntaxError:
		return None

	source_declarations = chunk_ast_nodes_by_path(
		filter_ast_node(source_tree, VariableDeclaration)
	)

	target_declarations = chunk_ast_nodes_by_path(
		filter_ast_node(target_tree, VariableDeclaration)
	)

	target_var_names = [
		declarator.name
		for path, declarations in target_declarations
		for declaration in declarations
		for declarator in declaration.declarators
	]

	source_references = chunk_ast_nodes_by_path(
		filter_ast_node(source_tree, MemberReference)
	)

	target_references = chunk_ast_nodes_by_path(
		filter_ast_node(target_tree, MemberReference)
	)

	renamed_var_names = []
	matches = []

	# Find all the semantically equalivalent variables that have different
	# names. For the variable declarations a and b to be semantically
	# equalivalent:
	#
	# - They must have be in the same semantic section of the file
	# - They must have the same references and assignments

	# Iterate over every pair of chunks, including the pair of the same chunk.
	for i in range(len(source_declarations)):
		source_decn_path, source_decn_nodes = source_declarations[i]
		for j in range(i, len(target_declarations)):
			target_decn_path, target_decn_nodes = target_declarations[j]

			# 1. Make sure paths would be equal if source_node (a) and
			# target_node (b) were the same.

			# 1a. Must be in the same part of the file
			# First get all previously discovered renamed variables
			prev_matches_here = [
				(old, new, attrs)
				for path, v in matches
				for old, new, attrs in v
				if path in (source_decn_path, target_decn_path)
			]

			# Now make sure the paths are equal (ignoring the previously found
			# renamed variables and the declarations themselves)
			if not ast_paths_equal(
				source_decn_path,
				target_decn_path,
				matches=prev_matches_here
			):
				continue

			for k in range(len(source_decn_nodes)):
				source_decn_node = source_decn_nodes[k]
				for m in range(len(source_decn_node.declarators)):
					source_declarator = source_decn_node.declarators[m]
					# 1b. Must have different names
					if source_declarator.name in target_var_names:
						continue

					sr = [
						(path, ref)
						for path, refs in source_references
						for ref in refs
						if ref.member == source_declarator.name
					]

					for n in range(k, len(target_decn_nodes)):
						target_decn_node = target_decn_nodes[n]
						for q in range(len(target_decn_node.declarators)):
							target_declarator = target_decn_node.declarators[q]
							# 1c. All attributes except the name should be equal
							declarator_match = (source_declarator, target_declarator, 'name')
							if not ast_nodes_equal(
								source_declarator,
								target_declarator,
								recursive=True,
								matches=prev_matches_here + [declarator_match]
							):
								continue

							# 1d. Must have the same references
							if not _have_same_references(
								source_declarator.name,
								target_declarator.name,
								source_references,
								target_references,
								prev_matches_here
							):
								continue

							tr = [
								(path, ref)
								for path, refs in target_references
								for ref in refs
								if ref.member == target_declarator.name
							]

							# (sr and tr have the same length, so we can do this safely)
							for i in range(len(sr)):
								source_ref_path, source_ref_node = sr[i]
								target_ref_path, target_ref_node = tr[i]
								ref_match = (source_ref_node, target_ref_node, 'name')

								for path, mappings in matches:
									if path == target_ref_path:
										mappings.append(ref_match)

								else:
									matches.append((target_ref_path, [ref_match]))

							# 1d. (The function cannot return multiple mappings
							# containing the same variable at the same path)
							other_renames_at_path = False
							never_used = True
							for path, mappings in renamed_var_names:
								if path != target_decn_path:
									continue

								for old, new in mappings:
									if old == source_declarator.name or new == target_declarator.name:
										never_used = False
										break

								if not never_used:
									break

								mappings.append((source_declarator.name, target_declarator.name))
								other_renames_at_path = True
								break

							if not never_used:
								continue

							if not other_renames_at_path:
								mappings = [(source_declarator.name, target_declarator.name)]
								renamed_var_names.append((target_decn_path, mappings))

							for path, v in matches:
								if path == target_decn_path:
									for old, new, _ in v:
										if old == source_declarator:
											raise JavaAnalyzationError(f'duplicate declarator in source: {old}')

										if new == target_declarator:
											raise JavaAnalyzationError(f'duplicate declarator in target: {new}')

									v.append(declarator_match)

							else:
								matches.append((target_decn_path, [declarator_match]))

	return renamed_var_names


def undo_variable_renames(code: str, renamed_var_names: List[Tuple]) -> str:
	lines = code.split('\n')
	tokens = tokenize(code)

	for token in tokens:
		if not isinstance(token, Identifier):
			continue

		for path, mappings in renamed_var_names:
			for old, new in mappings:
				if token.value == new:
					line, col = token.position
					line -= 1
					col -= 1
					lines[line] = lines[line][:col] + old + lines[line][col + len(new):]

	return '\n'.join(lines)


def undo_renames(repo: git.Repo) -> None:
	# Since mint does not parse commits into objects, the repo needs to be a #
	# gitpython repo

	commit1 = repo.commit('HEAD')
	commit2 = None  # working tree
	diff_index = commit1.diff(commit2)

	updated_count = 0

	# For each file changed (in reverse, with indices, so we can remove elements
	# in the for loop)
	for diff in diff_index:
		# Only process modified files (no new, deleted, ... files)
		if diff.change_type != 'M':
			continue

		# Only process Java files; leave everything else unchanged
		if not diff.a_path.endswith('.java'):
			continue

		source = get_blob(repo, commit1, diff.a_path)
		target = get_blob(repo, commit2, diff.b_path)

		try:
			renamed_variables = get_renamed_variables(source, target)
		except JavaAnalyzationError as e:
			raise Exception(f'{e} [{diff.a_path} -> {diff.b_path}]')

		if renamed_variables is not None:
			updated_target = undo_variable_renames(target, renamed_variables)
			p = os.path.join(repo.working_tree_dir, diff.a_path)
			with open(p, 'w') as f:
				f.write(updated_target)

			updated_count += 1

	print(f'+ Updated {updated_count} files')



import pytest

from mint.repo import Repo


@pytest.fixture(autouse=True)
def fs(mocker) -> None:
	mocker.patch('mint.command.subprocess.os.path.exists', return_value=True)
	mocker.patch('mint.command.subprocess.os.path.isfile', return_value=False)


@pytest.fixture(autouse=True)
def subprocess(mocker) -> None:
	mocker.patch('mint.command.subprocess')


class TestRepo:
	def test_clone_returns_repo_with_same_path(self):
		repo = Repo.clone('bar.git', 'bar')
		assert repo.path == 'bar'



import os
import shutil
import subprocess
import tempfile
from typing import List

import pytest


class RunParams:
	def __init__(
		self,
		versions: List[str],
		repo_path: str,
		undo_renamed_vars: bool
	) -> None:

		self.versions = versions
		self.repo_path = repo_path
		self.undo_renamed_vars = undo_renamed_vars


def create_command(
	versions: List[str],
	mappings: str,
	repo_path: str,
	undo_renamed_vars: bool
) -> List[str]:

	command = ['pipenv', 'run', 'start', '-p', repo_path, '--mappings', mappings]
	if undo_renamed_vars:
		command.append('-u')

	command.extend(versions)

	return command


def _run(versions: List[str], mappings: str, undo_renamed_vars: bool) -> None:
	with tempfile.TemporaryDirectory(prefix='shulkr') as repo_path:
		try:
			command = create_command(versions, mappings, repo_path, undo_renamed_vars)
			p = subprocess.run(command, stderr=subprocess.PIPE)

		except KeyboardInterrupt as e:
			if os.path.exists(repo_path):
				shutil.rmtree(repo_path)

			raise e

		if p.returncode != 0:
			if os.path.exists(repo_path):
				shutil.rmtree(repo_path)

			raise Exception(p.stderr.decode())

		yield RunParams(versions, repo_path, undo_renamed_vars)


@pytest.fixture(
	scope='session',
	params=[
		# Testing every combination of mappings to undo_variable_renames will
		# take too long, so just mix and match
		(['1.17.1', '1.18'], 'mojang', False),
	]
)
def run_mojang(request):
	versions, mappings, undo_variable_renames = request.param
	yield from _run(versions, mappings, undo_variable_renames)


@pytest.fixture(
	scope='session',
	params=[
		# Testing every combination of mappings to undo_variable_renames will
		# take too long, so just mix and match
		(['1.17.1', '1.18'], 'yarn', True),
	]
)
def run_yarn(request):
	versions, mappings, undo_variable_renames = request.param
	yield from _run(versions, mappings, undo_variable_renames)



from mint.command import GitCommandError

from shulkr.config import config_exists
from shulkr.repo import get_repo


def _repo_has_commits() -> bool:
	try:
		get_repo().git.rev_parse('HEAD')
		return True

	except GitCommandError:
		return False


def is_compatible() -> bool:
	"""
	Check if the current repo is compatible with the current shulkr version.
	"""

	repo = get_repo()
	return config_exists(repo.path) or not _repo_has_commits()



import os
import sys
from typing import List

import click
from minecraft.version import (
	NoSuchVersionError,
	Version,
	load_manifest
)

from shulkr.compatibility import is_compatible
from shulkr.config import init_config
from shulkr.gitignore import ensure_gitignore_exists
from shulkr.repo import init_repo
from shulkr.version import create_version, get_latest_generated_version


def run(
	versions: List[str],
	mappings: str,
	repo_path: str,
	message_template: str,
	tags: bool,
	undo_renamed_vars: bool
) -> None:

	load_manifest()

	full_repo_path = os.path.join(
		os.getcwd(),
		repo_path
	)

	init_output = not init_repo(full_repo_path)

	if not is_compatible():
		click.secho(
			'This repo is not compatible with the current version of shulkr - ' +
			'please create a new repo or downgrade shulkr.',
			err=True,
			fg='yellow'
		)
		sys.exit(4)

	init_output = not init_config(
		full_repo_path,
		mappings,
		message_template,
		tags,
		undo_renamed_vars
	) or init_output
	init_output = not ensure_gitignore_exists() or init_output

	# If we printed anything in the initialization step, print a newline
	if init_output:
		click.echo()

	try:
		resolved_versions = Version.patterns(
			versions,
			latest_in_repo=get_latest_generated_version()
		)

	except NoSuchVersionError as e:
		click.secho(e, err=True, fg='red')
		sys.exit(1)

	if len(resolved_versions) == 0:
		click.secho('No versions selected', color='yellow')
		sys.exit(0)

	if resolved_versions[0] < get_latest_generated_version():
		click.secho(
			'The latest version in the repo is ' +
			get_latest_generated_version().id +
			', but you selected ' +
			resolved_versions[0].id +
			'. Please select a version that is newer than the latest ' +
			'version in the repo.',
			err=True,
			fg='red'
		)
		sys.exit(3)

	for i, version in enumerate(resolved_versions):
		create_version(version)

		# Print line between the output of generating each version
		if i < len(resolved_versions) - 1:
			click.echo()



from mint.repo import Repo
from minecraft.version import Version

from shulkr.version import get_latest_generated_version


def test_get_latest_generated_version_with_repo_with_two_versions_returns_newer_version(repo: Repo):
	# Add two tagged commits (Minecraft versions)
	repo.git.commit(message='1.17', allow_empty=True)
	repo.git.tag('1.17')

	repo.git.commit(message='1.18', allow_empty=True)
	repo.git.tag('1.18')

	# The latest generated version from HEAD should be 1.18
	assert get_latest_generated_version() == Version.of('1.18')


def test_get_latest_generated_version_with_repo_with_two_versions_after_checking_out_older_version_returns_newer_version(repo: Repo):
	# Add two tagged commits (Minecraft versions)
	repo.git.commit(message='1.17', allow_empty=True)
	repo.git.tag('1.17')

	repo.git.commit(message='1.18', allow_empty=True)
	repo.git.tag('1.18')

	# Get the name of the current (only) branch
	orig_branch = repo.git.rev_parse('HEAD', abbrev_ref=True)

	# Now, check out 1.17 (this used to cause get_latest_generated_version to
	# return 1.17 instead of 1.18)
	repo.git.checkout('1.17')

	# Return to 1.18, without checking it out directly (important)
	repo.git.checkout(orig_branch)

	# The latest generated version from HEAD should be 1.18
	assert get_latest_generated_version() == Version.of('1.18')



from __future__ import annotations
import os

import git

from mint.command import GitCommand


class NoSuchRepoError(Exception):
	def __init__(self, path: str, *args: object) -> None:
		super().__init__(*args)

		self.path = path

	def __str__(self) -> str:
		return f'{self.path} is not a git repo'


class Repo:
	"""
	Git repo

	Creating a Repo object:
		# Use an existing repo
		repo = Repo(PATH)

		# Create an empty repo
		repo = Repo.init(PATH)

		# Clone a repo
		repo = Repo.clone(REMOTE_URL, DESTINATION)

	Using a Repo object:
		repo.git.commit('src', message='Some commit')  # or m='Some commit'
	"""

	def __init__(self, path: str, check_path=True) -> None:
		if check_path:
			Repo._ensure_repo_path_is_valid(path)

		self.path = path
		self.git = GitCommand(path)

	def to_gitpython(self) -> git.Repo:
		return git.Repo(self.path)

	@staticmethod
	def _ensure_repo_path_is_valid(path: str):
		if not os.path.exists(path):
			raise FileNotFoundError(path)

		# Make sure the path is a directory or a symlink
		if os.path.isfile(path):
			raise NotADirectoryError(path)

		git_dir = os.path.join(path, '.git')
		if not os.path.exists(git_dir) or os.path.isfile(git_dir):
			raise NoSuchRepoError(path)

	@staticmethod
	def init(path: str, **kwargs) -> Repo:
		if not os.path.exists(path):
			os.mkdir(path)

		repo = Repo(path, check_path=False)
		repo.git.init(**kwargs)
		return repo

	@staticmethod
	def clone(remote: str, dest: str, **kwargs) -> Repo:
		git = GitCommand()
		git.clone(remote, dest, **kwargs)
		return Repo(dest)



from glob import glob
import os.path
from setuptools import find_packages, setup
from typing import List


def package_files(directory: str) -> List[str]:
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


script_dir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(script_dir, 'README.md'), 'r') as f:
	README = f.read()

classifiers = [
	'Development Status :: 2 - Pre-Alpha',
	'Environment :: Console',
	'Intended Audience :: Developers',
	'Intended Audience :: Science/Research',
	'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
	'Operating System :: OS Independent',
	'Programming Language :: Python',
	'Programming Language :: Python :: 3',
	'Topic :: Games/Entertainment',
	'Topic :: Scientific/Engineering',
	'Topic :: Software Development :: Version Control :: Git',
	'Topic :: Utilities'
]

setup(
	name='shulkr',
	version='0.6.1',
	description='Decompile multiple versions of Minecraft with a single command (for research)',
	long_description=README,
	long_description_content_type='text/markdown',
	url='https://github.com/clabe45/shulkr',
	author='Caleb Sacks',
	license='GPLv3',
	classifiers=classifiers,
    keywords=['minecraft', 'git', 'decompile', 'game'],
    packages=find_packages(exclude=['tests']),
    py_modules=['java', 'minecraft', 'mint', 'shulkr'],
	# include_package_data=True,
	# package_data={'': package_files('shulkr/DecompilerMC')},
	install_requires=['gitpython', 'javalang', 'unidiff', 'requests', 'toml'],
	entry_points={
        'console_scripts': ['shulkr=shulkr.__main__:main']
    }
)


from mint.repo import Repo
import keepachangelog


CHANGELOG_PATH = 'docs/changelog.md'


def current_version():
	changes = keepachangelog.to_dict(CHANGELOG_PATH)
	return sorted(changes.keys())[-1]


def main():
	repo = Repo('.')

	old_version = current_version()

	# Release to get predicted version
	keepachangelog.release(CHANGELOG_PATH)
	predicted_version = current_version()

	# Confirm version
	user_input = input('New version [{}]: '.format(predicted_version))
	new_version = user_input or predicted_version

	# Undo temporary release
	repo.git.restore(CHANGELOG_PATH)

	# Release with confirmed version
	keepachangelog.release(CHANGELOG_PATH, new_version)

	# Update setup.py
	with open('setup.py', 'r') as setuppy:
		setuppy_code = setuppy.read()
	new_setuppy_code = setuppy_code.replace(old_version, new_version)
	with open('setup.py', 'w') as setuppy:
		setuppy.write(new_setuppy_code)

	# Commit to git
	commit_message = f'chore(*): release version {new_version}\n\nBump version {old_version} → {new_version}'
	repo.git.commit(CHANGELOG_PATH, 'setup.py', message=commit_message)
	repo.git.tag(f'v{new_version}', annotate=True, message=f'version {new_version}')

	print(f'Bumped version {old_version} → {new_version}')


if __name__ == '__main__':
	main()



from mint.command import GitCommand
import pytest


@pytest.fixture
def git(repo):
	return GitCommand(repo.path)


class TestCommand:
	def test_getattr_returns_git_output(self, git):
		# Make one commit so git knows which branch is checked out
		git.commit(message='dummy commit', allow_empty=True)

		# Now, make sure the result of 'git status' is correct
		branch = git.rev_parse('HEAD', abbrev_ref=True)
		assert git.status() == f'On branch {branch}\nnothing to commit, working tree clean'



import os

import click

from shulkr.repo import get_repo


def _gitignore_path() -> str:
	repo = get_repo()

	return os.path.join(repo.path, '.gitignore')


def _create_gitignore() -> None:
	click.echo('Creating gitignore')

	repo = get_repo()

	with open(_gitignore_path(), 'w+') as gitignore:
		to_ignore = ['.yarn', '.DecompilerMC']
		gitignore.write('\n'.join(to_ignore) + '\n')

	repo.git.add('.gitignore')
	repo.git.commit(message='add .gitignore')


def ensure_gitignore_exists() -> bool:
	"""
	Create and commit a .gitignore file if one does not exist

	Returns:
		bool: True if a .gitignore file was found. False if one was created.
	"""

	if not os.path.isfile(_gitignore_path()):
		_create_gitignore()
		return False

	else:
		return True



import os

from shulkr.gitignore import ensure_gitignore_exists


def test_ensure_gitignore_exists_opens_gitignore_for_writing_if_it_does_not_exist(mocker, empty_repo):
	mocker.patch('shulkr.gitignore.click')
	mocker.patch('shulkr.gitignore.os.path.isfile', return_value=False)
	open_ = mocker.patch('shulkr.gitignore.open')

	ensure_gitignore_exists()

	gitignore_path = os.path.join(empty_repo.path, '.gitignore')
	open_.assert_called_once_with(gitignore_path, 'w+')


def test_ensure_gitignore_exists_does_not_open_gitignore_if_it_does_exist(mocker, empty_repo):
	mocker.patch('shulkr.gitignore.click')
	mocker.patch('shulkr.gitignore.os.path.isfile', return_value=True)
	open_ = mocker.patch('shulkr.gitignore.open')

	ensure_gitignore_exists()

	open_.assert_not_called()


def test_ensure_gitignore_exists_returns_false_if_it_does_not_exist(mocker, empty_repo):
	mocker.patch('shulkr.gitignore.click')
	mocker.patch('shulkr.gitignore.os.path.isfile', return_value=False)
	mocker.patch('shulkr.gitignore.open')

	assert not ensure_gitignore_exists()


def test_ensure_gitignore_exists_returns_true_if_it_does_exist(mocker, empty_repo):
	mocker.patch('shulkr.gitignore.click')
	mocker.patch('shulkr.gitignore.os.path.isfile', return_value=True)
	mocker.patch('shulkr.gitignore.open')

	assert ensure_gitignore_exists()
